// SDCARD Data Logger with an FIFO buffer
// The FIFO is used to eliminate the SDcard's Write Latency issue (WL 5-250ms)
// The FIFO uses two semaphores to synchronize between 2 tasks:
//    SensorsTask - is the Producer (writes ADC etc. to the FIFO)
//    SdcardTask  - is the Consumer (writes data Records from the FIFO to the SDcard)
// IDE 1.6.10+, SdFat beta latest, STM32duino core latest
// Based on the BillG's SdFat Fifo example
// Maple Mini and other mods by Pito 10/2016
// Bonus: ADC samples FIR Filter implemented
// License: MIT
// Provided as-is, No warranties of any kind, Use at your own risk :)
// c Pito 10/2016
//
// Sketch uses 33,292 bytes (27%) of program storage space. Maximum is 122,880 bytes.
// Global variables use 20,088 bytes of dynamic memory.


#include <SPI.h>
#include <MapleFreeRTOS821.h>
#include <SdFat.h>
#include "fir_filter30hz_200hzsampling.c"

#define BOARD_LED_PIN PB1

// Interval between Data Samples Records in units of 1ms (the FreeRtos tick)
const uint16_t SampleIntervalTicks = 5;  // 5 = 200Hz sampling rate

//------------------------------------------------------------------------------
// SD file definitions
const uint8_t sdChipSelect = PB12;
SdFat sd;
SdFile file;

//------------------------------------------------------------------------------
// FIFO buffer definitions
// Size of the FIFO in number of Sample Data Records
const size_t FIFO_SIZE = 150;

// Data type for a single Sample Data Record
struct FifoItem_t {
  uint32_t usec;
  int16_t value1;
  int16_t value2;
  int16_t value3;
  int16_t value4;
  int16_t value5;
  int16_t value6;
  int16_t value7;
  int16_t value8;
  uint16_t errors;
  uint16_t a;
  uint16_t b;
};

// Array of the Data Sample Records
FifoItem_t fifoArray[FIFO_SIZE];

// Semaphore's definitions
// We do not have this function in this FreeRtos version - a workaround
#define uxSemaphoreGetCount(xSemaphore) uxQueueMessagesWaiting((QueueHandle_t)(xSemaphore))

// Count of the actual Data Records items in the FIFO
SemaphoreHandle_t fifoData;

// Count of free Data Records Space in the FIFO
SemaphoreHandle_t fifoFree;

// FIR filters definition
SampleFilter f[8];

//------------------------------------------------------------------------------
// Producer Task - it samples 8 ADC inputs etc. into the FIFO
static void vSensorsTask(void *pvParameters) {
  // Index of a Record to be filled into FIFO
  size_t fifoHead = 0;

  // Count of overrun errors
  uint16_t overrunerrors = 0;
  
  // Initialise the ticks variable with the current Rtos time
  TickType_t ticks = xTaskGetTickCount();

  while (1) {
    // Wait until the Time for the next Data Sample
    
    vTaskDelayUntil(&ticks, SampleIntervalTicks);

    // Check out if there is a free space in the FIFO

    if (xSemaphoreTake(fifoFree, 0) != pdTRUE) {
    // FIFO is full - indicates a missed Data Sample Record(s)
      overrunerrors++;
      continue;  // Wait on the next data sample
    }

    FifoItem_t* p = &fifoArray[fifoHead];

    // Record the actual Sample period start
    p->usec = micros();

    // Record the ADC data read from the Sensors into the FIFO
    // Sampling of 8 ADC with 71taps fir filtering takes 285usecs
    p->value1 = analogRead(PA0);
    p->value2 = analogRead(PA1);
    p->value3 = analogRead(PA2);
    p->value4 = analogRead(PA3);
    p->value5 = analogRead(PA4);
    p->value6 = analogRead(PA5);
    p->value7 = analogRead(PA6);
    p->value8 = analogRead(PA7);
    
    SampleFilter_put(&f[0], (int) p->value1);
    p->value1 = SampleFilter_get(&f[0]);
    SampleFilter_put(&f[1], (int) p->value2);
    p->value2 = SampleFilter_get(&f[1]);    
    SampleFilter_put(&f[2], (int) p->value3);
    p->value3 = SampleFilter_get(&f[2]);    
    SampleFilter_put(&f[3], (int) p->value4);
    p->value4 = SampleFilter_get(&f[3]);    
    SampleFilter_put(&f[4], (int) p->value5);
    p->value5 = SampleFilter_get(&f[4]);    
    SampleFilter_put(&f[5], (int) p->value6);
    p->value6 = SampleFilter_get(&f[5]);    
    SampleFilter_put(&f[6], (int) p->value7);
    p->value7 = SampleFilter_get(&f[6]);    
    SampleFilter_put(&f[7], (int) p->value8);
    p->value8 = SampleFilter_get(&f[7]);

    // Record the FIFO overruns for this sample
    p->errors = overrunerrors;
    overrunerrors = 0; // clear for the next sample

    // Advance the FIFO index to the next free record
    fifoHead = fifoHead < (FIFO_SIZE - 1) ? fifoHead + 1 : 0;

    // Record the actual values of counting semaphores (for demo only)
    p->a = uxSemaphoreGetCount(fifoFree);
    p->b = uxSemaphoreGetCount(fifoData);
    
    // Signal to SDcardTask the new Data Record is waiting in the FIFO
    xSemaphoreGive(fifoData);
    }
}

//------------------------------------------------------------------------------
// Consumer Task - write Data to the SDcard task
static void vSdcardTask(void *pvParameters) {
  // FIFO index for a record to be written onto the SDcard
  size_t fifoTail = 0;

  // Time in micros of the last measurement
  uint32_t last = 0;

  while(1) {
    // Wait for the next available Data Record
    if (xSemaphoreTake(fifoData, portMAX_DELAY) != pdTRUE) {
      // None data available, wait for next data record
      continue;
    }

    // Set the actual pointer to FIFO array item
    FifoItem_t* p = &fifoArray[fifoTail];

    // Print out the Data Record values into the CSV file

    if (last) {
      file.print(p->usec - last);
    } else {
      file.write("NA");
    }
    last = p->usec;
        
    file.write(',');
    file.print(p->value1);
    file.write(',');
    file.print(p->value2);
    file.write(',');
    file.print(p->value3);
    file.write(',');
    file.print(p->value4);
    file.write(',');
    file.print(p->value5);
    file.write(',');
    file.print(p->value6);
    file.write(',');
    file.print(p->value7);
    file.write(',');
    file.print(p->value8);
    file.write(',');
    file.print(p->errors);
    file.write(',');
    file.print(p->a);
    file.write(',');
    file.println(p->b);

    // Advance the FIFO index
    fifoTail = fifoTail < (FIFO_SIZE - 1) ? fifoTail + 1 : 0;
    
    // Release the Record
    // Signal to the SensorsTask a Data Record has been written onto the SDcard
    xSemaphoreGive(fifoFree);

    // Check for an End of the Measurement
    if (Serial1.available()) {
      // Close the file to ensure the data is saved correctly
      file.close();
      Serial1.println(F("Done.."));
      while(1);
    }
  }
}

 // Perform an LED blinking every 200msecs
 static void vLEDFlashTask( void * pvParameters ) { 
 portTickType xLastWakeTime; 
 const portTickType xPeriod = 200;
     // Initialise the xLastWakeTime variable with the current time. 
     xLastWakeTime = xTaskGetTickCount(); 
     for( ;; ) //
     { 
         // Wait for the next blink cycle. 
         vTaskDelayUntil( &xLastWakeTime, xPeriod );
         // Perform the LED blink 
         digitalWrite(BOARD_LED_PIN, HIGH);
         vTaskDelay(25); // 25msecs long blink
         digitalWrite(BOARD_LED_PIN, LOW); 
     } 
}

//------------------------------------------------------------------------------
void setup() {
  // Tasks creation statuses
  portBASE_TYPE s1, s2, s3;

  Serial1.begin(115200);

  for(int i = 0; i < 8; i++) {
    SampleFilter_init(&f[i]);
  }

  // Start of a Measurement
  Serial1.println(F("Type any character to begin"));
  while(!Serial1.available());

  // Open a CSV file
  if (!sd.begin(sdChipSelect)|| !file.open("DATA.CSV", O_CREAT | O_WRITE | O_TRUNC)) {
    Serial1.println(F("SDcard problem.."));
    while(1);
  }
  
  // Initialize fifoData semaphore to "None data records available"
  fifoData = xSemaphoreCreateCounting(FIFO_SIZE, 0);

  // Initialize fifoFree semaphore to "FIFO_SIZE free records available"
  fifoFree = xSemaphoreCreateCounting(FIFO_SIZE, FIFO_SIZE);

  // Create Sensors capture task at priority 2 (higher)
  s1 = xTaskCreate(vSensorsTask, NULL, configMINIMAL_STACK_SIZE + 200, NULL, tskIDLE_PRIORITY + 2, NULL);

  // Create SD write task at priority 1 (lower)
  s2 = xTaskCreate(vSdcardTask, NULL, configMINIMAL_STACK_SIZE + 200, NULL, tskIDLE_PRIORITY + 1, NULL);

  // Create LED blinking task  
  s3 = xTaskCreate(vLEDFlashTask, NULL, configMINIMAL_STACK_SIZE, NULL, tskIDLE_PRIORITY + 0, NULL);

  // Check for Task's and Semaphore's creation errors
  if (fifoData == NULL || fifoFree == NULL || s1 != pdPASS || s2 != pdPASS ) {
    Serial1.println(F("Task or Semaphore creation problem.."));
    while(1);
  }
  
  // Throw away serial input
  while (Serial1.available()) Serial1.read();
  Serial1.println(F("Type any character to end"));

  pinMode(BOARD_LED_PIN, OUTPUT);

  // Print out Data Record Headers into the CSV file
  PrntHeaders();

  // Start the task scheduler - here the process starts
  vTaskStartScheduler();

  // In case of a scheduler error
  Serial1.println(F("Die - insufficient heap memory?"));
  while(1);
}
//------------------------------------------------------------------------------

void loop() {
  // not used here
}

void PrntHeaders () {
    // Print Data Record Headers into the CSV file
    file.print(F("Time_us"));
    file.write(',');
    file.print(F("ADC0"));
    file.write(',');
    file.print(F("ADC1"));
    file.write(',');
    file.print(F("ADC2"));
    file.write(',');
    file.print(F("ADC3"));
    file.write(',');
    file.print(F("ADC4"));
    file.write(',');
    file.print(F("ADC5"));
    file.write(',');
    file.print(F("ADC6"));
    file.write(',');
    file.print(F("ADC7"));
    file.write(',');
    file.print(F("Overruns"));
    file.write(',');
    file.print(F("FifoFree"));
    file.write(',');
    file.println(F("FifoData"));
}

